
'''
EM3D : Frequency domain Maxwell equation.

   This module is meant to solve 

    (curl v, curl u) + (v, e u) 
           - (v, n x (1/mu curl u)_s = i w (jext, u) 

    , where e = - epsion * w^2 - i * sigma * w, w is frequency,
    and  mu, epislon, and simga is regular EM parameters.

    ( , ) is volume integral and ( , )_s is surface intengral
    on non-essential boundaries.

    exp(-i w t) is assumed.

 *sigma
  Domain:   
     EM3D_Anisotropic : tensor dielectric
     EM3D_Vac         : scalar dielectric
     EM3D_ExtJ        : external current
     EM3D_Div         : div J = 0 constraints (add Lagrange multiplier)

  Boundary:
     EM3D_PEC         : Perfect electric conductor
     EM3D_PMC         : Perfect magnetic conductor
     EM3D_H           : Mangetic field boundary
     EM3D_SurfJ       : Surface current
     EM3D_Port        : TE, TEM, Coax port
     EM3D_E           : Electric field
     EM3D_Continuity  : Continuitiy

  Pair:
     EM3D_Floquet     : Periodic boundary condition
'''
from petram.phys.vtable import VtableElement, Vtable
import numpy as np
import traceback

from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model import Phys
from petram.phys.common.em_base import EMPhysModule
from petram.phys.em3d.em3d_base import EM3D_Bdry
from petram.phys.em3d.em3d_vac import EM3D_Vac

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3DModel')


class EM3D_DefDomain(EM3D_Vac):
    can_delete = False
    nlterms = []
    # vt  = Vtable(data1)
    # do not use vtable here, since we want to use
    # vtable defined in EM3D_Vac in add_bf_conttribution

    def __init__(self, **kwargs):
        super(EM3D_DefDomain, self).__init__(**kwargs)

    def panel1_param(self):
        return [['Default Domain (Vac)',   "eps_r=1, mu_r=1, sigma=0",  2, {}], ]

    def get_panel1_value(self):
        return ["eps_r=1, mu_r=1, sigma=0", ]

    def import_panel1_value(self, v):
        pass

    def panel1_tip(self):
        return None

    def get_possible_domain(self):
        return []

    def get_possible_child(self):
        return self.parent.get_possible_domain()


data2 = (('label1', VtableElement(None,
                                  guilabel='Default Bdry (PMC)',
                                  default="Ht = 0",
                                  tip="this is a natural BC")),)


class EM3D_DefBdry(EM3D_Bdry):
    can_delete = False
    is_essential = False
    nlterms = []
    vt = Vtable(data2)

    def __init__(self, **kwargs):
        super(EM3D_DefBdry, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM3D_DefBdry, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = ['remaining']
        return v
    '''   
    def panel1_param(self):
        return [['Default Bdry (PMC)',   "Ht = 0",  2, {}],]

    def get_panel1_value(self):
        return None

    def import_panel1_value(self, v):
        pass
    
    def panel1_tip(self):
        return None
    '''

    def get_possible_bdry(self):
        return []


class EM3D_DefPair(Pair, Phys):
    can_delete = False
    is_essential = False

    def __init__(self, **kwargs):
        super(EM3D_DefPair, self).__init__(**kwargs)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM3D_DefPair, self).attribute_set(v)
        v['sel_readonly'] = False
        v['sel_index'] = []
        return v

    def get_possible_pair(self):
        return []


class EM3D(EMPhysModule):
    der_var_base = ['Bx', 'By', 'Bz']
    der_var_vec = ['B']
    geom_dim = 3

    def __init__(self, **kwargs):
        super(EM3D, self).__init__(**kwargs)
        self['Domain'] = EM3D_DefDomain()
        self['Boundary'] = EM3D_DefBdry()
        self['Pair'] = EM3D_DefPair()

    @property
    def dep_vars(self):
        '''
        list of dependent variables, for example.
           [E]      
           [E, psi]
        '''
        ret = ['E']
        if self._has_div_constraint():
            ret = ['E', 'psi']
        else:
            ret = ['E']
        return [x + self.dep_vars_suffix for x in ret]

    @property
    def dep_vars0(self):
        '''
        list of dependent variables, for example.
           [E]      
           [E, psi]
        '''
        ret = ['E']
        if self._has_div_constraint():
            ret = ['E', 'psi']
        else:
            ret = ['E']
        return [x + self.dep_vars_suffix for x in ret]

    @property
    def dep_vars_base(self):
        ret = ['E']
        if self._has_div_constraint():
            ret = ['E', 'psi']
        else:
            ret = ['E']
        return ret

    def get_fec_type(self, idx):
        values = ['ND', 'H1']
        return values[idx]

    def get_fec(self):
        v = self.dep_vars
        if len(v) == 1:  # normal case
            return [(v[0], 'ND_FECollection')]
        else:            # with divergence constraints
            return [(v[0], 'ND_FECollection'),
                    (v[1], 'H1_FECollection'), ]

    def _has_div_constraint(self):
        from petram.phys.em3d.em3d_div import EM3D_Div
        for mm in self['Domain'].iter_enabled():
            if isinstance(mm, EM3D_Div):
                return True
        return False

    def attribute_set(self, v):
        v = super(EM3D, self).attribute_set(v)
        v["element"] = 'ND_FECollection'
        v["freq_txt"] = "1.0e9"
        v["ndim"] = 3
        v["ind_vars"] = 'x, y, z'
        v["dep_vars_suffix"] = ''
        return v

    def panel1_param(self):
        panels = super(EM3D, self).panel1_param()
        a, b = self.get_var_suffix_var_name_panel()
        panels.extend([  # self.make_param_panel('freq',  self.freq_txt),
            ["independent vars.", self.ind_vars, 0, {}],
            a,
            ["dep. vars.", ','.join(self.dep_vars), 2, {}],
            ["ns vars.", ','.join(EM3D.der_var_base), 2, {}], ])

        return panels

    def get_panel1_value(self):
        names = ', '.join(self.dep_vars)
        names2 = ', '.join(list(self.get_default_ns()))
        val = super(EM3D, self).get_panel1_value()
        val.extend([  # self.freq_txt,
            self.ind_vars, self.dep_vars_suffix,
            names, names2, ])
        return val

    def import_panel1_value(self, v):
        v = super(EM3D, self).import_panel1_value(v)
        # self.freq_txt = str(v[0])
        self.ind_vars = str(v[0])
        self.dep_vars_suffix = str(v[1])

    def get_possible_bdry(self):
        if EM3D._possible_constraints is None:
            self._set_possible_constraints('em3d')
        bdrs = super(EM3D, self).get_possible_bdry()
        return EM3D._possible_constraints['bdry'] + bdrs

    def get_possible_domain(self):
        if EM3D._possible_constraints is None:
            self._set_possible_constraints('em3d')

        doms = super(EM3D, self).get_possible_domain()
        return EM3D._possible_constraints['domain'] + doms

    '''
    def get_possible_bdry(self):
        from .em3d_pec         import EM3D_PEC
        from .em3d_pmc         import EM3D_PMC
        from .em3d_h           import EM3D_H
        from .em3d_surfj       import EM3D_SurfJ
        from .em3d_port        import EM3D_Port
        from .em3d_portarray        import EM3D_PortArray
        from .em3d_e           import EM3D_E
        from .em3d_cont        import EM3D_Continuity
        from .em3d_z           import EM3D_Impedance
        bdrs = super(EM3D, self).get_possible_bdry()
        return [EM3D_PEC, EM3D_Port, EM3D_PortArray, EM3D_E, EM3D_SurfJ, 
                EM3D_H, EM3D_PMC, EM3D_Impedance, EM3D_Continuity]+bdrs
    
    def get_possible_domain(self):
        from .em3d_anisotropic import EM3D_Anisotropic
        from .em3d_vac         import EM3D_Vac
        from .em3d_extj        import EM3D_ExtJ
        from .em3d_div         import EM3D_Div
        
        doms = super(EM3D, self).get_possible_domain()
        
        return [EM3D_Vac, EM3D_Anisotropic, EM3D_ExtJ, EM3D_Div] + doms
    '''

    def get_possible_edge(self):
        return []

    '''
    def get_possible_pair(self):

        from .em3d_floquet     import EM3D_Floquet

        return [EM3D_Floquet]
    '''

    def get_possible_pair(self):
        if EM3D._possible_constraints is None:
            self._set_possible_constraints()
        return EM3D._possible_constraints['pair']

    '''
    def get_possible_point(self):
        return []
    '''

    def add_variables(self, v, name, solr, soli=None):
        from petram.helper.variables import add_coordinates
        from petram.helper.variables import add_scalar
        from petram.helper.variables import add_components
        from petram.helper.variables import add_component_expression as addc_expression
        from petram.helper.variables import add_expression
        from petram.helper.variables import add_surf_normals
        from petram.helper.variables import add_constant

        from petram.helper.eval_deriv import eval_curl

        v = super(EM3D, self).add_variables(v, name, solr, soli)

        freq, omega = self.get_freq_omega()

        def evalB(gfr, gfi=None):
            gfr, gfi, extra = eval_curl(gfr, gfi)
            gfi /= omega   # real B
            gfr /= -omega  # imag B
            # flipping gfi and gfr so that it returns
            # -i * (-gfr + i gfi) = gfi + i gfr
            return gfi, gfr, extra

        ind_vars = [x.strip()
                    for x in self.ind_vars.split(',') if x.strip() != '']

        suffix = self.dep_vars_suffix

        from petram.helper.variables import TestVariable
        # v['debug_test'] =  TestVariable()

        add_coordinates(v, ind_vars)
        add_surf_normals(v, ind_vars)

        from petram.phys.phys_const import mu0, epsilon0, q0

        if name.startswith('E'):

            add_components(v, 'E', suffix, ind_vars, solr, soli)
            add_components(v, 'B', suffix, ind_vars, solr, soli,
                           deriv=evalB)
            add_expression(v, 'normE', suffix, ind_vars,
                           '(conj(Ex)*Ex + conj(Ey)*Ey +conj(Ez)*Ez)**(0.5)',
                           ['E'])

            add_expression(v, 'normB', suffix, ind_vars,
                           '(conj(Bx)*Bx + conj(By)*By + conj(Bz)*Bz)**(0.5)',
                           ['B'])

            # Poynting Flux
            addc_expression(v, 'Poy', suffix, ind_vars,
                            '(conj(Ey)*Bz - conj(Ez)*By)/mu0',
                            ['B', 'E'], 0)
            addc_expression(v, 'Poy', suffix, ind_vars,
                            '(conj(Ez)*Bx - conj(Ex)*Bz)/mu0',
                            ['B', 'E'], 1)
            addc_expression(v, 'Poy', suffix, ind_vars,
                            '(conj(Ex)*By - conj(Ey)*Bx)/mu0',
                            ['B', 'E'], 2)

            # e = - epsion * w^2 - i * sigma * w
            # Jd : displacement current  = -i omega* e0 er E
            addc_expression(v, 'Jd', suffix, ind_vars,
                            '(-1j*(dot(epsilonr, E))*freq*2*pi*e0)[0]',
                            ['epsilonr', 'E', 'freq'],  0)
            addc_expression(v, 'Jd', suffix, ind_vars,
                            '(-1j*(dot(epsilonr, E))*freq*2*pi*e0)[1]',
                            ['epsilonr', 'E', 'freq'], 1)
            addc_expression(v, 'Jd', suffix, ind_vars,
                            '(-1j*(dot(epsilonr, E))*freq*2*pi*e0)[2]',
                            ['epsilonr', 'E', 'freq'], 2)
            # Ji : induced current = sigma *E
            addc_expression(v, 'Ji', suffix, ind_vars,
                            '(dot(sigma, E))[0]',
                            ['sigma', 'E'], 0)
            addc_expression(v, 'Ji', suffix, ind_vars,
                            '(dot(sigma, E))[1]',
                            ['sigma', 'E'], 1)
            addc_expression(v, 'Ji', suffix, ind_vars,
                            '(dot(sigma, E))[2]',
                            ['sigma', 'E'], 2)
            # Jp : polarization current (Jp = -i omega* e0 (er - 1) E
            addc_expression(v, 'Jp', suffix, ind_vars,
                            '(-1j*(dot(epsilonr, E) - E)*freq*2*pi*e0)[0]',
                            ['epsilonr', 'E', 'freq'], 0)
            addc_expression(v, 'Jp', suffix, ind_vars,
                            '(-1j*(dot(epsilonr, E) - E)*freq*2*pi*e0)[1]',
                            ['epsilonr', 'E', 'freq'], 1)
            addc_expression(v, 'Jp', suffix, ind_vars,
                            '(-1j*(dot(epsilonr, E) - E)*freq*2*pi*e0)[2]',
                            ['epsilonr', 'E', 'freq'], 2)

            # Js : surface current (n x B / mu)
            addc_expression(v, 'Js', suffix, ind_vars,
                            '(cross([nx, ny, nz], inv(mur).dot(B))/mu0)[0]',
                            ['nx', 'ny', 'nz', 'B', 'mur', 'mu0'], 0)
            addc_expression(v, 'Js', suffix, ind_vars,
                            '(cross([nx, ny, nz], inv(mur).dot(B))/mu0)[1]',
                            ['nx', 'ny', 'nz', 'B', 'mur', 'mu0'], 1)
            addc_expression(v, 'Js', suffix, ind_vars,
                            '(cross([nx, ny, nz], inv(mur).dot(B))/mu0)[2]',
                            ['nx', 'ny', 'nz', 'B', 'mur', 'mu0'], 2)

        elif name.startswith('psi'):
            add_scalar(v, 'psi', suffix, ind_vars, solr, soli)

        # collect all definition from children
        # for mm in self.walk():
        #    if not mm.enabled: continue
        #    if mm is self: continue
        #    mm.add_domain_variables(v, name, suffix, ind_vars,
        #                            solr, soli)
        #    mm.add_bdr_variables(v, name, suffix, ind_vars,
        #                            solr, soli)

        return v

    def get_fes_for_dep(self, unknown_name, soldict):
        keys = soldict.keys()
        for k in keys:
            if unknown_name.startswith('psi'):
                if k.startswith('psi'):
                    break
            else:
                if k.startswith('E'):
                    break
        sol = soldict[k]
        solr = sol[0]
        soli = sol[1] if len(sol) > 1 else None
        return solr, soli



'''

  evaulate derivative of Nedelec element using
  DiscreteLinearOperator

'''
from petram.mfem_config import use_parallel
import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('eval_deriv')

if use_parallel:
   import mfem.par as mfem
   FiniteElementSpace = mfem.ParFiniteElementSpace
   DiscreteLinearOperator = mfem.ParDiscreteLinearOperator
   GridFunction = mfem.ParGridFunction
else:
   import mfem.ser as mfem
   FiniteElementSpace = mfem.FiniteElementSpace
   DiscreteLinearOperator = mfem.DiscreteLinearOperator
   GridFunction = mfem.GridFunction

def eval_curl(gfr, gfi = None):
    fes = gfr.FESpace()
    ordering = fes.GetOrdering()
    mesh = fes.GetMesh()
    vdim = 1
    sdim = mesh.SpaceDimension()
    p = fes.GetOrder(0)
    rt_coll = mfem.RT_FECollection(p-1, sdim)

    rts = FiniteElementSpace(mesh,  rt_coll, vdim, ordering)

    
    curl = DiscreteLinearOperator(fes, rts)
    itp = mfem.CurlInterpolator()
    curl.AddDomainInterpolator(itp)
    curl.Assemble();
    curl.Finalize();

    br   = GridFunction(rts)    
    curl.Mult(gfr,  br)
    if gfi is not None:
       bi = GridFunction(rts)           
       curl.Mult(gfi,  bi)
    else:
       bi = None
    ### needs to return rts to prevent rts to be collected.
    return br, bi,  rts





'''
   anistropic region:
   note that MFEM3.3 does not support matrix form of coefficeint
   for curl-curl. Needs next update.
'''
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.coefficient import MCoeff
from petram.mfem_config import use_parallel
import numpy as np
from scipy.linalg import inv
from petram.phys.phys_model import MatrixPhysCoefficient, PhysCoefficient, PhysConstant, PhysMatrixConstant
from petram.phys.em3d.em3d_base import EM3D_Bdry, EM3D_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_Anisotropic')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('epsilonr', VtableElement('epsilonr', type='complex',
                                   guilabel='epsilonr',
                                   suffix=[('x', 'y', 'z'), ('x', 'y', 'z')],
                                   default=np.eye(3, 3),
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur', type='complex',
                              guilabel='mur',
                              suffix=[('x', 'y', 'z'), ('x', 'y', 'z')],
                              default=np.eye(3, 3),
                              tip="relative permeability")),
        ('sigma', VtableElement('sigma', type='complex',
                                guilabel='sigma',
                                suffix=[('x', 'y', 'z'), ('x', 'y', 'z')],
                                default=np.zeros((3, 3)),
                                tip="contuctivity")),)


def Epsilon_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # - omega^2 * epsilon0 * epsilonr
    fac = -epsilon0 * omega * omega * cnorm
    coeff = MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def Sigma_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    # v = - 1j * self.omega * v
    fac = - 1j * omega * cnorm
    coeff = MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def Mu_Coeff(exprs, ind_vars, l, g, omega, cnorm):
    fac = mu0/cnorm
    coeff = MCoeff(3, exprs, ind_vars, l, g, return_complex=True, scale=fac)
    return coeff


def domain_constraints():
    return [EM3D_Anisotropic]


class EM3D_Anisotropic(EM3D_Domain):
    allow_custom_intorder = True
    vt = Vtable(data)
    #nlterms = ['epsilonr']

    def get_possible_child(self):
        from .em3d_pml import EM3D_LinearPML
        return [EM3D_LinearPML]

    def has_bf_contribution(self, kfes):
        if kfes == 0:
            return True
        else:
            return False

    def get_coeffs(self, real=True):
        freq, omega = self.get_root_phys().get_freq_omega()
        cnorm = self.get_root_phys().get_coeff_norm()

        e, m, s = self.vt.make_value_or_expression(self)

        ind_vars = self.get_root_phys().ind_vars
        l = self._local_ns
        g = self._global_ns
        coeff1 = Epsilon_Coeff(e, ind_vars, l, g, omega, cnorm)
        coeff2 = Mu_Coeff(m, ind_vars, l, g, omega, cnorm)
        coeff3 = Sigma_Coeff(s, ind_vars, l, g, omega, cnorm)

        return coeff1, coeff2, coeff3

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        if kfes != 0:
            return
        if real:
            dprint1("Add BF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add BF contribution(imag)" + str(self._sel_index))

        freq, omega = self.get_root_phys().get_freq_omega()

        coeff1, coeff2, coeff3 = self.get_coeffs()
        self.set_integrator_realimag_mode(real)

        if self.has_pml():
            coeff1 = self.make_PML_coeff(coeff1)
            coeff2 = self.make_PML_coeff(coeff2)
            coeff3 = self.make_PML_coeff(coeff3)
        coeff2 = coeff2.inv()

        if self.allow_custom_intorder and self.add_intorder != 0:
            fes = a.FESpace()
            geom = fes.GetFE(0).GetGeomType()
            order = fes.GetFE(0).GetOrder()
            isPK = (fes.GetFE(0).Space() == mfem.FunctionSpace.Pk)
            orderw = fes.GetElementTransformation(0).OrderW()
            curlcurl_order = order*2 - 2 if isPK else order*2
            mass_order = orderw + 2*order
            curlcurl_order += self.add_intorder
            mass_order += self.add_intorder

            dprint1("Debug: custom int order. Increment = " +
                    str(self.add_intorder))
            dprint1("  FE order: " + str(order))
            dprint1("  OrderW: " + str(orderw))
            dprint1("  CurlCurlOrder: " + str(curlcurl_order))
            dprint1("  FEMassOrder: " + str(mass_order))

            cc_ir = mfem.IntRules.Get(geom, curlcurl_order)
            ms_ir = mfem.IntRules.Get(geom, mass_order)
        else:
            cc_ir = None
            ms_ir = None

        self.add_integrator(engine, 'epsilonr', coeff1,
                            a.AddDomainIntegrator,
                            mfem.VectorFEMassIntegrator,
                            ir=ms_ir)

        if coeff2 is not None:
            self.add_integrator(engine, 'mur', coeff2,
                                a.AddDomainIntegrator,
                                mfem.CurlCurlIntegrator,
                                ir=cc_ir)
            #coeff2 = self.restrict_coeff(coeff2, engine)
            # a.AddDomainIntegrator(mfem.CurlCurlIntegrator(coeff2))
        else:
            dprint1("No contrinbution from curlcurl")

        self.add_integrator(engine, 'sigma', coeff3,
                            a.AddDomainIntegrator,
                            mfem.VectorFEMassIntegrator,
                            ir=ms_ir)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        if len(self._sel_index) == 0:
            return

        e, m, s = self.vt.make_value_or_expression(self)

        self.do_add_matrix_expr(v, suffix, ind_vars, 'epsilonr', e)
        self.do_add_matrix_expr(v, suffix, ind_vars, 'mur', m)
        self.do_add_matrix_expr(v, suffix, ind_vars, 'sigma', s)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')



'''
   external current source
'''
from petram.phys.phys_model import PhysCoefficient
from petram.phys.phys_model import VectorPhysCoefficient
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.coefficient import VCoeff, SCoeff
from petram.mfem_config import use_parallel
import numpy as np

from petram.phys.em2da.em2da_base import EM2Da_Bdry, EM2Da_Domain

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2Da_H')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

data = (('ht', VtableElement('ht', type='complex',
                             guilabel='H',
                             suffix=('r', 'phi', 'z'),
                             default=[0, 0, 0],
                             tip="boundary H field.")),)


class rHp(VectorPhysCoefficient):  # i \omega r Jext_t
    def __init__(self, *args, **kwargs):
        self.omega = kwargs.pop('omega', 1.0)
        super(rHp, self).__init__(*args, **kwargs)

    def Eval(self, V, T, ip):
        nor = mfem.Vector(2)
        mfem.CalcOrtho(T.Jacobian(), nor)
        tmp = nor.GetDataArray()
        tmp = np.array([tmp[0], 0, tmp[1]])
        self.nor = tmp/np.linalg.norm(tmp)

        return VectorPhysCoefficient.Eval(self, V, T, ip)

    def EvalValue(self, x):
        v = super(rHp, self).EvalValue(x)
        v = np.cross(self.nor, v)

        v = np.array((v[0], v[2]))
        v = -1j * self.omega * v * x[0]
        if self.real:
            return v.real
        else:
            return v.imag


class Ht(PhysCoefficient):  # i \omega Jext_phi
    def __init__(self, *args, **kwargs):
        self.omega = kwargs.pop('omega', 1.0)
        super(Ht, self).__init__(*args, **kwargs)

    def Eval(self, T, ip):
        nor = mfem.Vector(2)
        mfem.CalcOrtho(T.Jacobian(), nor)
        tmp = nor.GetDataArray()
        tmp = np.array([tmp[0], 0, tmp[1]])
        self.nor = tmp/np.linalg.norm(tmp)
        return PhysCoefficient.Eval(self, T, ip)

    def EvalValue(self, x):
        v = super(Ht, self).EvalValue(x)
        v = np.cross(self.nor, v)

        v = -1j * self.omega * v[1]

        if self.real:
            return v.real
        else:
            return v.imag


def bdry_constraints():
    return [EM2Da_H]


class EM2Da_H(EM2Da_Bdry):
    is_essential = False
    vt = Vtable(data)

    def has_lf_contribution(self, kfes=0):
        if kfes > 2:
            return False
        return True

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if kfes < 2:
            if real:
                dprint1("Add LF contribution(real)" + str(self._sel_index))
            else:
                dprint1("Add LF contribution(imag)" + str(self._sel_index))
            freq, omega = self.get_root_phys().get_freq_omega()
            f_name = self.vt.make_value_or_expression(self)

            print(kfes, f_name)
            if kfes == 0:
                coeff1 = rHp(2, f_name[0],  self.get_root_phys().ind_vars,
                             self._local_ns, self._global_ns,
                             real=real, omega=omega)
                self.add_integrator(engine, 'ht', coeff1,
                                    b.AddBoundaryIntegrator,
                                    mfem.VectorFEDomainLFIntegrator)
            else:
                coeff1 = Ht(f_name[0],  self.get_root_phys().ind_vars,
                            self._local_ns, self._global_ns,
                            real=real, omega=omega)
                self.add_integrator(engine, 'htor', coeff1,
                                    b.AddBoundaryIntegrator,
                                    mfem.DomainLFIntegrator)

        else:
            assert False, "should not come here"



'''
   1D port boundary condition
'''
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
from petram.phys.phys_const import mu0, epsilon0

from mfem.common.mpi_debug import nicePrint
import sys
import numpy as np

from petram.model import Bdry
from petram.phys.phys_model import Phys
from petram.phys.em1d.em1d_base import EM1D_Bdry, EM1D_Domain

from petram.helper.geom import connect_pairs2
from petram.helper.geom import find_circle_center_radius

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM1D_Port')


if use_parallel:
    import mfem.par as mfem
    '''
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   '''
else:
    import mfem.ser as mfem


data = (('inc_amp', VtableElement('inc_amp', type='complex',
                                  guilabel='inc. amp.',
                                  default=(1.0, 0.0),
                                  suffix=('y', 'z'),
                                  no_func=True,
                                  tip="amplitude of incoming wave")),
        ('inc_phase', VtableElement('inc_phase', type='float',
                                    guilabel='inc. phase (deg)',
                                    default=0.0,
                                    no_func=True,
                                    tip="phase of incoming wave")),
        ('epsilonr', VtableElement('epsilonr', type='complex',
                                   guilabel='epsilonr',
                                   default=1.0,
                                   no_func=True,
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur', type='complex',
                              guilabel='mur',
                              default=1.0,
                              no_func=True,
                              tip="relative permeability")),
        ('ky', VtableElement('ky', type='float',
                             guilabel='ky',
                             default=0.,
                             no_func=True,
                             tip="wave number` in the y direction")),
        ('kz', VtableElement('kz', type='float',
                             guilabel='kz',
                             default=0.0,
                             no_func=True,
                             tip="wave number in the z direction")),)


def bdry_constraints():
    return [EM1D_Port]


class E_port(mfem.PyCoefficient):
    def __init__(self, bdry, real=True, amp=(1.0, 0.0),  eps=1.0,
                 mur=1.0, ky=0.0, kz=0.0, direction="y", normalize=False):

        mfem.PyCoefficient.__init__(self)
        self.real = real
        freq, omega = bdry.get_root_phys().get_freq_omega()

        k = omega*np.sqrt(eps*epsilon0 * mur*mu0)
        kc2 = k**2 - ky**2 - kz**2
        if kc2 < 0:
            kc2 = complex(kc2)
            #raise ValueError('Mode does not propagate')
        beta = np.sqrt(kc2)
        dprint1("propagation constant:" + str(beta))
        Ey, Ez = amp
        Ex = -(Ey*ky + Ez*kz)/beta
        if normalize:
            E_norm = np.sqrt(Ex*np.conj(Ex) + Ey*np.conj(Ey) + Ez*np.conj(Ez))
            Ex = Ex/E_norm
            Ey = Ey/E_norm
            Ez = Ez/E_norm
        if direction == "y":
            self.ret = Ey
        else:
            self.ret = Ez
        self.ret = complex(self.ret)

    def EvalValue(self, x):
        if self.real:
            return self.ret.real
        else:
            return self.ret.imag


class jwH_port(mfem.PyCoefficient):
    def __init__(self, bdry, real=True, amp=(1.0, 0.0),  eps=1.0,
                 mur=1.0, ky=0.0, kz=0.0, direction="y", normalize=False):
        mfem.PyCoefficient.__init__(self)
        self.real = real
        freq, omega = bdry.get_root_phys().get_freq_omega()
        cnorm = bdry.get_root_phys().get_coeff_norm()

        k = omega*np.sqrt(eps*epsilon0 * mur*mu0)
        kc2 = k**2 - ky**2 - kz**2

        if kc2 < 0:
            dprint1('Mode does not propagate !!!')
            kc2 = complex(kc2)
            #raise ValueError('Mode does not propagate')

        norm = -bdry.norm  # norm is INWARD propagation
        beta = np.sqrt(kc2) * norm
        dprint1("propagation constant:" + str(beta))
        Ey, Ez = amp
        Ex = -(Ey*ky + Ez*kz)/beta

        if normalize:
            E_norm = np.sqrt(Ex*np.conj(Ex) + Ey*np.conj(Ey) + Ez*np.conj(Ez))
            Ex = Ex/E_norm
            Ey = Ey/E_norm
            Ez = Ez/E_norm
        Hy = 1j*(-beta*Ez + kz*Ex)/mu0/mur
        Hz = 1j*(beta*Ey - ky*Ex)/mu0/mur

        # return n x H (n is outward vector
        if direction == "y":
            self.ret = -Hz * (-bdry.norm)
        else:
            self.ret = Hy * (-bdry.norm)
        self.ret = complex(self.ret)*cnorm

    def EvalValue(self, x):
        if self.real:
            return self.ret.real
        else:
            return self.ret.imag


class EM1D_Port(EM1D_Bdry):
    extra_diagnostic_print = True
    vt = Vtable(data)

    def __init__(self, inc_amp='1, 0', inc_phase='0', port_idx=1):
        super(EM1D_Port, self).__init__(inc_amp=inc_amp,
                                        inc_phase=inc_phase,
                                        port_idx=port_idx)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM1D_Port, self).attribute_set(v)
        v['port_idx'] = 1
        v['sel_readonly'] = False
        v['sel_index'] = []
        v['isTimeDependent_RHS'] = True
        self.vt.attribute_set(v)
        return v

    def panel1_param(self):
        return ([["port id", str(self.port_idx), 0, {}], ] +
                self.vt.panel_param(self))

    def get_panel1_value(self):
        return ([str(self.port_idx), ] +
                self.vt.get_panel_value(self))

    def import_panel1_value(self, v):
        self.port_idx = v[0]
        self.vt.import_panel_value(self, v[1:])

    def panel4_param(self):
        ll = super(EM1D_Port, self).panel4_param()
        ll.append(['Varying (in time/for loop) RHS', False, 3, {"text": ""}])
        return ll

    def panel4_tip(self):
        return None

    def import_panel4_value(self, value):
        super(EM1D_Port, self).import_panel4_value(value[:-1])
        self.isTimeDependent_RHS = value[-1]

    def get_panel4_value(self):
        value = super(EM1D_Port, self).get_panel4_value()
        value.append(self.isTimeDependent_RHS)
        return value

    def verify_setting(self):
        if self.isTimeDependent_RHS:
            flag = True
        else:
            flag = False
        return flag, 'Varying RHS is not set', 'This potntially causes an error with PortScan. Set it Time/NL Dep. panel '

    def update_param(self):
        self.update_inc_amp_phase()

    def update_inc_amp_phase(self):
        '''
        set parameter
        '''
        try:
            self.vt.preprocess_params(self)
            inc_amp, inc_phase, eps, mur, ky, kz = self.vt.make_value_or_expression(
                self)
            self.inc_amp = inc_amp
            self.inc_phase = inc_phase
        except:
            raise ValueError("Cannot evaluate amplitude/phase to float number")

    def preprocess_params(self, engine):
        # find normal (outward) vector...
        mesh = engine.get_emesh(mm=self)
        fespace = engine.fespaces[self.get_root_phys().dep_vars[1]]

        nbe = mesh.GetNBE()

        BdrPtx = []
        for i in range(nbe):
            iv = mesh.GetBdrElementVertices(i)
            BdrPtx.append(mesh.GetVertexArray(iv[0])[0])

        ibe = np.array([i for i in range(nbe)
                        if mesh.GetBdrElement(i).GetAttribute() ==
                        self._sel_index[0]])
        iv = mesh.GetBdrElementVertices(ibe[0])
        ptx = mesh.GetVertexArray(iv[0])[0]
        if ptx == BdrPtx[-1]:
            self.norm = 1.
        else:
            self.norm = -1.

    def has_lf_contribution(self, kfes):
        self.vt.preprocess_params(self)
        inc_amp, inc_phase, eps, mur, ky, kz = self.vt.make_value_or_expression(
            self)
        if kfes == 1:
            return inc_amp[0] != 0
        elif kfes == 2:
            return inc_amp[1] != 0
        else:
            return False

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if kfes == 0:
            return

        txt = ["Ex in ", "Ey in ", "Ez in "][kfes]
        if real:
            dprint1("Add LF contribution(real) " + txt + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + txt + str(self._sel_index))

        self.vt.preprocess_params(self)
        inc_amp, inc_phase, eps, mur, ky, kz = self.vt.make_value_or_expression(
            self)

        inc_wave = np.array(inc_amp) * np.exp(1j*inc_phase/180.*np.pi)
        d = "y" if kfes == 1 else "z"

        # note for the right hand side, we multiple -1 to ampulitude

        coeff = jwH_port(self, real=real, amp=inc_wave,  eps=eps,
                         mur=mur, ky=ky, kz=kz,  direction=d)
        self.add_integrator(engine, 'inc_amp', coeff,
                            b.AddBoundaryIntegrator,
                            mfem.BoundaryLFIntegrator)

    def has_extra_DoF(self, kfes):
        if kfes == 0:
            return False
        else:
            return True

    def get_extra_NDoF(self):
        return 1

    def extra_DoF_name2(self, kfes=0):
        '''
        default DoF name
        '''
        if kfes == 0:
            assert False, "should not come here"
        elif kfes == 1:
            return self.get_root_phys().dep_vars[1]+"_port"+str(self.port_idx)
        elif kfes == 2:
            return self.get_root_phys().dep_vars[2]+"_port"+str(self.port_idx)

    def get_probes(self):
        return [self.get_root_phys().dep_vars[1]+"_port"+str(self.port_idx),
                self.get_root_phys().dep_vars[2]+"_port"+str(self.port_idx), ]

    def postprocess_extra(self, sol, flag, sol_extra):
        assert False, "Is it used?"
        name = self.name()+'_' + str(self.port_idx)
        sol_extra[name] = sol.toarray()

    def add_extra_contribution(self, engine, **kwargs):

        from mfem.common.chypre import LF2PyVec

        kfes = kwargs.pop('kfes', 0)
        if kfes == 0:
            return
        dprint1("Add Extra contribution" + str(self._sel_index))

        self.vt.preprocess_params(self)
        inc_amp, inc_phase, eps, mur, ky, kz = self.vt.make_value_or_expression(
            self)

        fes = engine.get_fes(self.get_root_phys(), kfes)
        d = "y" if kfes == 1 else "z"
        inc_amp0 = (1, 0) if kfes == 1 else (0, 1)
        lf1 = engine.new_lf(fes)
        Ht = jwH_port(self, real=True, amp=inc_amp0,  eps=eps,
                      mur=mur, ky=ky, kz=kz,  direction=d)  # , normalize=True)
        Ht = self.restrict_coeff(Ht, engine)
        intg = mfem.BoundaryLFIntegrator(Ht)
        lf1.AddBoundaryIntegrator(intg)
        lf1.Assemble()

        lf1i = engine.new_lf(fes)
        Ht = jwH_port(self, real=False, amp=inc_amp0,  eps=eps,
                      mur=mur, ky=ky, kz=kz,  direction=d)  # , normalize=True)
        Ht = self.restrict_coeff(Ht, engine)
        intg = mfem.BoundaryLFIntegrator(Ht)
        lf1i.AddBoundaryIntegrator(intg)
        lf1i.Assemble()

        from mfem.common.chypre import LF2PyVec, PyVec2PyMat, Array2PyVec, IdentityPyMat

        v1 = LF2PyVec(lf1, lf1i)
        #v1 *= -1

        lf2 = engine.new_lf(fes)
        Et = E_port(self, real=True, amp=inc_amp0,  eps=eps,
                    mur=mur, ky=ky, kz=kz,  direction=d)  # , normalize=True)

        Et = self.restrict_coeff(Et, engine)
        intg = mfem.DomainLFIntegrator(Et)
        lf2.AddBoundaryIntegrator(intg)
        lf2.Assemble()

        x = engine.new_gf(fes)
        x.Assign(0.0)
        arr = self.get_restriction_array(engine)
        x.ProjectBdrCoefficient(Et,  arr)

        t4 = np.array([[inc_amp[kfes-1]*np.exp(1j*inc_phase/180.*np.pi)]])

        weight = mfem.InnerProduct(engine.x2X(x), engine.b2B(lf2))

        v2 = LF2PyVec(lf2, None, horizontal=True)
        x = LF2PyVec(x, None)

        v2 *= 1/weight

        # convert to a matrix form
        v1 = PyVec2PyMat(v1)
        v2 = PyVec2PyMat(v2.transpose())
        t4 = Array2PyVec(t4)
        t3 = IdentityPyMat(1, diag=-1)

        v2 = v2.transpose()

        # return (None, v2, t3, t4, True)
        return (v1, v2, t3, t4, True)



'''

   3D port array boundary condition


   TE Mode
    Expression are based on Microwave Engineering p122 - p123.
    Note that it consists from two terms
       1)  \int dS W \dot n \times iwH (VectorFETangentIntegrator does this)
       2)  an weighting to evaulate mode amplutude from the E field
           on a boundary
   TEM Mode
       E is parallel to the periodic edge
       Mode number is ignored (of course)

   About sign of port phasing.
       positive phasing means the incoming wave appears later (phase delay)
       at the port

    2020 7/20  generated from portarray
'''
from petram.phys.common.rf_port_geometry import (analyze_geom_norm_ref,
                                                 analyze_rect_geom,
                                                 analyze_coax_geom,
                                                 analyze_circular_geom)
from petram.phys.em3d.em3d_base import EM3D_Bdry, EM3D_Domain
from petram.phys.phys_model import Phys
from petram.model import Bdry, Pair
from petram.phys.vtable import VtableElement, Vtable
from mfem.common.chypre import (LF2PyVec,
                                PyVec2PyMat,
                                Array2PyVec,
                                IdentityPyMat,
                                HStackPyVec)
from petram.mfem_config import use_parallel

from itertools import cycle
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_PortArray')

if use_parallel:
    import mfem.par as mfem
    from mfem.common.mpi_debug import nicePrint
    from petram.helper.mpi_recipes import *

else:
    import mfem.ser as mfem
    nicePrint = dprint1


data = (('inc_amp', VtableElement('inc_amp',
                                  type='array',
                                  guilabel='incoming amp',
                                  default=0.0,
                                  tip="amplitude of incoming wave")),
        ('inc_phase', VtableElement('inc_phase',
                                    type='float',
                                    guilabel='incoming phase (deg)',
                                    default=0.0,
                                    tip="phase of incoming wave")),
        ('inc_dphase', VtableElement('inc_dphase',
                                     type='float',
                                     guilabel='phase progression (deg)',
                                     default=0.0,
                                     tip="phase progression of incoming wave")),
        ('epsilonr', VtableElement('epsilonr',
                                   type='complex',
                                   guilabel='epsilonr',
                                   default=1.0,
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur',
                              type='complex',
                              guilabel='mur',
                              default=1.0,
                              tip="relative permeability")),)


def bdry_constraints():
    return [EM3D_PortArray]


class EM3D_PortArray(EM3D_Bdry):
    extra_diagnostic_print = True    
    vt = Vtable(data)

    def __init__(self, mode='TE', mn='0,1', inc_amp='1',
                 inc_phase='0', inc_dphase='90', port_idx=1):
        super(EM3D_PortArray, self).__init__(mode=mode,
                                             mn=mn,
                                             inc_amp=inc_amp,
                                             inc_phase=inc_phase,
                                             inc_dphase=inc_dphase,
                                             port_idx=port_idx)
        Phys.__init__(self)

    def extra_DoF_name(self):
        p1 = int(self.port_idx)
        p2 = len(self._sel_index)
        txt = str(p1) + '_' + str(p2)
        return self.get_root_phys().dep_vars[0] + "_port_" + txt

    def get_probes(self):
        p1 = int(self.port_idx)
        p2 = len(self._sel_index)
        txt = str(p1) + '_' + str(p2)
        return [self.get_root_phys().dep_vars[0] + "_port_" + txt]

    def attribute_set(self, v):
        super(EM3D_PortArray, self).attribute_set(v)
        v['port_idx'] = 1
        v['mode'] = 'TE'
        v['mn'] = [1, 0]
        v['inc_amp_txt'] = '1.0'
        v['inc_phase_txt'] = '0.0'
        v['inc_dphase_txt'] = '90.0'
        v['inc_amp'] = 1.0
        v['inc_phase'] = 0.0
        v['inc_dphase'] = 0.0
        v['epsilonr'] = 1.0
        v['mur'] = 1.0
        v['sel_readonly'] = False
        v['sel_index'] = []
        v['isTimeDependent_RHS'] = True
        v['ref_pt'] = '1'        
        return v

    def panel1_param(self):
        return ([["port group id", str(self.port_idx), 0, {}],
                 ["mode", self.mode, 4, {"readonly": True,
                                         "choices": ["TE", "TEM", "Coax(TEM)"]}],
                 ["m/n", ','.join(str(x) for x in self.mn), 0, {}], 
                 ["ref. pt.", "", 0, {}],] +                
                self.vt.panel_param(self))

    def get_panel1_value(self):
        return ([str(self.port_idx),
                 self.mode, ','.join(str(x) for x in self.mn),
                 self.ref_pt] +
                self.vt.get_panel_value(self))

    def import_panel1_value(self, v):
        self.port_idx = v[0]
        self.mode = v[1]
        self.mn = [int(x) for x in v[2].split(',')]
        self.ref_pt = v[3]        
        self.vt.import_panel_value(self, v[4:])

    def panel4_param(self):
        ll = super(EM3D_PortArray, self).panel4_param()
        ll.append(['Varying (in time/for loop) RHS', False, 3, {"text": ""}])
        return ll

    def panel4_tip(self):
        return None

    def import_panel4_value(self, value):
        super(EM3D_PortArray, self).import_panel4_value(value[:-1])
        self.isTimeDependent_RHS = value[-1]

    def get_panel4_value(self):
        value = super(EM3D_PortArray, self).get_panel4_value()
        value.append(self.isTimeDependent_RHS)
        return value

    def verify_setting(self):
        if self.isTimeDependent_RHS:
            flag = True
        else:
            flag = False
        return flag, 'Varying RHS is not set', 'This potntially causes an error with PortScan. Set it Time/NL Dep. panel '

    def update_param(self):
        self.vt.preprocess_params(self)
        inc_amp, inc_phase, inc_dphase, eps, mur = self.vt.make_value_or_expression(
            self)

        '''
        self.update_inc_amp_phase()

    def update_inc_amp_phase(self):
        try:
            self.inc_amp = self.eval_phys_expr(
                str(self.inc_amp_txt),  'inc_amp')[0]
            self.inc_phase = self.eval_phys_expr(
                str(self.inc_phase_txt), 'inc_phase', chk_float = True)[0]
        except:
            raise ValueError("Cannot evaluate amplitude/phase to float number")
        '''

    def do_preprocess_params(self, engine):
        self.a = 0
        self.b = 0
        self.a_vec = 0
        self.b_vec = 0
        self.c = 0
        self.ctr = 0
        
        # find normal (outward) vector and reference point location.
        mesh = engine.get_emesh(mm=self)
        nbe = mesh.GetNBE()
        ibe = np.array([i for i in range(nbe)
                        if mesh.GetBdrElement(i).GetAttribute() ==
                        self._sel_index[0]])

        norm, rptx = analyze_geom_norm_ref(mesh, ibe, self.ref_pt)
        self.norm = norm

        dprint1("Normal Vector " + str(self.norm))
        dprint1("Ref. Point" + str(rptx))

        # find rectangular shape
        if str(self.mode).upper().strip() in ['TE', 'TM', 'TEM']:
            geom_data, vv = analyze_rect_geom(self, mesh, ibe, norm)
            self.a = geom_data["a"]
            self.a_vec = geom_data["a_vec"]
            self.b = geom_data["b"]
            self.b_vec = geom_data["b_vec"]
            self.c = geom_data["c"]
            dprint1("Long Edge  " + self.a.__repr__())
            dprint1("Long Edge Vec." + list(self.a_vec).__repr__())
            dprint1("Short Edge  " + self.b.__repr__())
            dprint1("Short Edge Vec." + list(self.b_vec).__repr__())
            
        elif self.mode == 'Coax(TEM)':
            geom_data, vv = analyze_coax_geom(mesh, ibe, norm, rptx)
            self.a = geom_data["a"]
            self.b = geom_data["b"]
            self.ctr = geom_data["ctr"]
            self.ax1 = geom_data["ax1"]
            self.ax2 = geom_data["ax2"]
            
        C_Et, C_jwHt = self.get_coeff_cls()

        self.vt.preprocess_params(self)
        inc_amp, inc_phase, inc_dphase, eps, mur = self.vt.make_value_or_expression(
            self)

        dprint1("E field pattern", eps, mur)
        Et = C_Et(3, self, real=True, eps=eps, mur=mur,
                  m=self.mn[0], n=self.mn[1])
        for p in vv:
            dprint1(p.__repr__() + ' : ' + Et.EvalValue(p).__repr__())
        dprint1("H field pattern")
        cnorm = self.get_root_phys().get_coeff_norm()
        Ht = C_jwHt(3, self, real=False, eps=eps, mur=mur, cnorm=cnorm,
                    m=self.mn[0], n=self.mn[1])
        for p in vv:
            dprint1(p.__repr__() + ' : ' + Ht.EvalValue(p).__repr__())

    def read_portparams(self):
        return (self.a, self.b, self.a_vec, self.b_vec, self.c, self.ctr)

    def set_portparams(self, params):
        self.a, self.b, self.a_vec, self.b_vec, self.c, self.ctr = params

    def preprocess_params(self, engine):
        sels = self._sel_index

        port_params = []
        for s in sels:
            self._sel_index = [s]
            self.do_preprocess_params(engine)
            params = self.read_portparams()
            port_params.append(params)

        self._port_params = port_params
        self._sel_index = sels

    def get_coeff_cls(self):
        from petram.phys.common.rf_portmode import get_portmode_coeff_cls

        return get_portmode_coeff_cls(self.mode)

    def has_lf_contribution(self, kfes):
        if kfes != 0:
            return False
        self.vt.preprocess_params(self)
        inc_amp, inc_phase, inc_dphase, eps, mur = self.vt.make_value_or_expression(
            self)

        return np.any(np.array(inc_amp) != 0)

    def _fixup_inc_amp(self, inc_amp):

        num_ports = len(self._sel_index)

        if len(inc_amp) == 1:
            inc_amp = [inc_amp[0]]*num_ports
        elif len(inc_amp) != num_ports:
            assert False, "Number of ports are not the same as number of incoming amplitude"
        else:
            pass  # looks good
        return inc_amp, num_ports

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if real:
            dprint1("Add LF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + str(self._sel_index))

        self.vt.preprocess_params(self)
        inc_amp, inc_phase, inc_dphase, eps, mur = self.vt.make_value_or_expression(
            self)

        C_Et, C_jwHt = self.get_coeff_cls()

        sels = self._sel_index

        from itertools import cycle
        dphase = cycle(np.atleast_1d(inc_dphase))

        inc_amp, num_ports = self._fixup_inc_amp(inc_amp)
        ph = inc_phase

        dprint1("Power, Phase: ", inc_amp, inc_phase)

        for kport in range(num_ports):
            sel = self._sel_index[kport]
            params = self._port_params[kport]
            inc_a = inc_amp[kport]

            self.set_portparams(params)

            inc_wave = inc_a * np.exp(1j * ph / 180. * np.pi)

            phase = np.angle(inc_wave) * 180 / np.pi
            amp = np.sqrt(np.abs(inc_wave))
            cnorm = self.get_root_phys().get_coeff_norm()

            Ht = C_jwHt(3, self, real=real, amp=amp, eps=eps, mur=mur, phase=phase, cnorm=cnorm,
                        m=self.mn[0], n=self.mn[1])
            Ht = self.restrict_coeff(Ht, engine, vec=True, idx=[sel])

            intg = mfem.VectorFEBoundaryTangentLFIntegrator(Ht)
            b.AddBoundaryIntegrator(intg)

            ph = ph + next(dphase)

    def has_extra_DoF(self, kfes):
        if kfes != 0:
            return False
        return True

    def get_extra_NDoF(self):
        return len(self._sel_index)
        # (in future) must be number of modes on this port...

    def is_extra_RHSonly(self):
        return True

    def postprocess_extra(self, sol, flag, sol_extra):
        name = self.name() + '_' + str(self.port_idx)
        sol_extra[name] = sol.toarray()

    def do_add_extra_contribution(
            self, engine, inc_amp, inc_phase, eps, mur, sel):
        C_Et, C_jwHt = self.get_coeff_cls()
        cnorm = self.get_root_phys().get_coeff_norm()

        fes = engine.get_fes(self.get_root_phys(), 0)

        lf1 = engine.new_lf(fes)
        Ht1 = C_jwHt(3, self, real=True, eps=eps, mur=mur, cnorm=cnorm,
                     m=self.mn[0], n=self.mn[1])
        Ht2 = self.restrict_coeff(Ht1, engine, vec=True, idx=[sel])
        intg = mfem.VectorFEBoundaryTangentLFIntegrator(Ht2)
        lf1.AddBoundaryIntegrator(intg)
        lf1.Assemble()
        lf1i = engine.new_lf(fes)
        Ht3 = C_jwHt(3, self, real=False, eps=eps, mur=mur, cnorm=cnorm,
                     m=self.mn[0], n=self.mn[1])
        Ht4 = self.restrict_coeff(Ht3, engine, vec=True, idx=[sel])
        intg = mfem.VectorFEBoundaryTangentLFIntegrator(Ht4)
        lf1i.AddBoundaryIntegrator(intg)
        lf1i.Assemble()

        lf2r = engine.new_lf(fes)
        Etr = C_Et(3, self, real=True, eps=eps, mur=mur,
                  m=self.mn[0], n=self.mn[1])
        Etr = self.restrict_coeff(Etr, engine, vec=True, idx=[sel])
        intg = mfem.VectorFEDomainLFIntegrator(Etr)
        lf2r.AddBoundaryIntegrator(intg)
        lf2r.Assemble()
        lf2i = engine.new_lf(fes)
        Eti = C_Et(3, self, real=False, eps=eps, mur=mur,
                   m=self.mn[0], n=self.mn[1])
        Eti = self.restrict_coeff(Eti, engine, vec=True, idx=[sel])
        intg = mfem.VectorFEDomainLFIntegrator(Eti)
        lf2i.AddBoundaryIntegrator(intg)
        lf2i.Assemble()

        arr = self.get_restriction_array(engine)

        xr = engine.new_gf(fes)
        xr.Assign(0.0)
        xr.ProjectBdrCoefficientTangent(Etr, arr)
        xi = engine.new_gf(fes)
        xi.Assign(0.0)
        xi.ProjectBdrCoefficientTangent(Eti, arr)

        et = engine.x2X(xr).GetDataArray() + \
                1j * engine.x2X(xi).GetDataArray()

        vec = engine.b2B(lf2r).GetDataArray() - \
                1j * engine.b2B(lf2i).GetDataArray()  # complex conjugate

        weight = np.sum(et.dot(vec))
        if use_parallel:
            weight = np.sum(allgather(weight))

        # t1
        t1 = LF2PyVec(lf1, lf1i)
        t1 *= 1


        # t2
        lf2i *= -1 # complex conjugate
        t2 = LF2PyVec(lf2r, lf2i, horizontal=True)
        t2 *= -1. / weight
        t2 = t2.transpose()

        # t4
        t4 = - np.array(
            [[np.sqrt(inc_amp) * np.exp(1j * inc_phase / 180. * np.pi)]])
        
        return (t1, t2, t4)

    def add_extra_contribution(self, engine, **kwargs):
        '''
        compute extra (multiplier) contribution

        Format of extar   (t2 is returnd as vertical(transposed) matrix)
        [M,  t1]   [  ]
        [      ] = [  ]
        [t2, t3]   [t4]

        and it returns if Lagurangian will be saved.
        '''
        dprint1("Add Extra contribution" + str(self._sel_index))

        self.vt.preprocess_params(self)
        inc_amp, inc_phase, inc_dphase, eps, mur = self.vt.make_value_or_expression(
            self)
        inc_amp, num_ports = self._fixup_inc_amp(inc_amp)
        ph = inc_phase
        dphase = cycle(np.atleast_1d(inc_dphase))

        dprint1("Power, Phase: ", inc_amp, inc_phase)

        t1_arr = []
        t2_arr = []
        t4 = []

        for kport in range(num_ports):
            sel = self._sel_index[kport]
            params = self._port_params[kport]
            inc_a = inc_amp[kport]

            self.set_portparams(params)
            # dprint1("phasing ", ph, sel)
            t1, t2, t4_1 = self.do_add_extra_contribution(
                engine, inc_a, ph, eps, mur, sel)

            t1_arr.append(t1)
            t2_arr.append(t2)

            t4.append(t4_1)
            ph = ph + next(dphase)

        t3 = IdentityPyMat(len(self._sel_index))
        t4 = Array2PyVec(np.array(t4))

        t1 = HStackPyVec(t1_arr)
        t2 = HStackPyVec(t2_arr)
        t2 = t2.transpose()

        #print(t1.shape, t2.shape, t3.shape, t4.shape)

        return (t1, t2, t3, t4, True)



from __future__ import print_function
from petram.phys.vtable import VtableElement, Vtable
from petram.phys.em3d.em3d_base import EM3D_Bdry, EM3D_Domain
from petram.phys.phys_model import Phys
from petram.model import Bdry
from petram.mfem_config import use_parallel
'''

   Port load. Define an external circuite attached to Port BC using
   external S-matrix

'''

import numpy as np
from numpy import pi

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_PortLoad')


if use_parallel:
    import mfem.par as mfem
    from mfem.common.mpi_debug import nicePrint
    '''
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   '''
else:
    import mfem.ser as mfem
    nicePrint = dprint1

data = (('port_s', VtableElement('port_s', type='array',
                                 guilabel='external S-mat',
                                 default="Smat",
                                 tip="S-matrix of external circuit attached to the ports")),
        ('port_id', VtableElement('port_id', type='array',
                                  guilabel='coupled ports',
                                  default="1, 2",
                                  tip="Port index of coupled ports. S-matrix size has to be larger than number of coupled ports")),
        ('ext_drive', VtableElement('ext_drive', type='array',
                                    guilabel='exteranl drive',
                                    default="",
                                    tip="External drive terms of S-matrix")),)


def bdry_constraints():
    return [EM3D_PortLoad]


class EM3D_PortLoad(EM3D_Bdry):
    extra_diagnostic_print = True
    vt = Vtable(data)

    def __init__(self):
        super(EM3D_PortLoad, self).__init__()

    def extra_DoF_name(self):
        return self.get_root_phys().dep_vars[0] + "_"+self.name().lower()

    def get_probes(self):
        return [self.get_root_phys().dep_vars[0] + "_"+self.name().lower()]

    def attribute_set(self, v):
        super(EM3D_PortLoad, self).attribute_set(v)
        v['isTimeDependent_RHS'] = True
        return v

    def panel1_param(self):
        return self.vt.panel_param(self)

    def get_panel1_value(self):
        return self.vt.get_panel_value(self)

    def import_panel1_value(self, v):
        self.vt.import_panel_value(self, v)

    def panel2_param(self):
        return [[None, "Auxiriary varialbe coupling is global",  2,   {}], ]

    def import_panel2_value(self, v):
        pass

    def get_panel2_value(self):
        return [None]

    def panel4_tip(self):
        return None

    def update_param(self):
        self.vt.preprocess_params(self)
        Smat, port_idx, ext_drive = self.vt.make_value_or_expression(self)

    def has_extra_DoF(self, kfes):
        if kfes != 0:
            return False
        return True

    def get_extra_NDoF(self):
        Smat, port_idx, ext_drive = self.vt.make_value_or_expression(self)

        l = len(port_idx)
        if ext_drive is not None:
            l = l + len(ext_drive)
        return l

    def postprocess_extra(self, sol, flag, sol_extra):
        name = self.name()
        sol_extra[name] = sol.toarray()

    def check_extra_update(self, mode):
        '''
        mode = 'B' or 'M'
        'M' return True, if M needs to be updated
        'B' return True, if B needs to be updated
        '''
        if self._update_flag:
            if mode == 'B':
                return self.isTimeDependent_RHS
            if mode == 'M':
                return self.isTimeDependent
        return False

    def _get_coupled_ports(self):
        from petram.phys.em3d.em3d_port import EM3D_Port

        root_bdr = self.get_root_phys()['Boundary']

        Smat, port_idx, ext_drive = self.vt.make_value_or_expression(self)
        # print(Smat, port_idx, ext_drive)
        target = []
        for mm in root_bdr.get_children():
            if not mm.enabled:
                continue
            if not isinstance(mm, EM3D_Port):
                continue

            idx = int(mm.port_idx)
            if idx in port_idx:
                k = np.where(port_idx == idx)[0][0]
                target.append((k, mm, mm.extra_DoF_name()))

        ret = sorted(target)
        return ret

    def add_extra_contribution(self, engine, **kwargs):
        '''
        Format of extar   (t2 is returnd as vertical(transposed) matrix)
        [M,  t1]   [  ]
        [      ] = [  ]
        [t2, t3]   [t4]
        and it returns if Lagurangian will be saved.
        '''
        dprint1("Add Extra contribution" + str(self._sel_index))
        from mfem.common.chypre import (LF2PyVec,
                                        PyVec2PyMat,
                                        Array2PyVec,
                                        IdentityPyMat,
                                        HStackPyVec)

        Smat, port_idx, ext_drive = self.vt.make_value_or_expression(self)
        ports = self. _get_coupled_ports()
        names2 = [x[2] for x in ports]

        l = self.get_extra_NDoF()
        if len(np.atleast_1d(Smat)) == 1:
            Smat = np.eye(l)*Smat

        Sright = Smat[:, len(names2):]
        l = self.get_extra_NDoF()

        vecs = []
        fes = engine.get_fes(self.get_root_phys(), 0)
        for x in ports:
            mm = x[1]
            C_Et, C_jwHt = mm.get_coeff_cls()
            cnorm = self.get_root_phys().get_coeff_norm()
            inc_amp, inc_phase, eps, mur = mm.vt.make_value_or_expression(mm)
            lf1 = engine.new_lf(fes)
            Ht1 = C_jwHt(3, mm, real=True, eps=eps, mur=mur, phase=180., cnorm=cnorm,
                         m=self.mn[0], n=self.mn[1])
            Ht2 = mm.restrict_coeff(Ht1, engine, vec=True, idx=mm._sel_index)
            intg = mfem.VectorFEBoundaryTangentLFIntegrator(Ht2)
            lf1.AddBoundaryIntegrator(intg)
            lf1.Assemble()
            lf1i = engine.new_lf(fes)
            Ht3 = C_jwHt(3, mm, real=False, eps=eps, mur=mur, phase=180., cnorm=cnorm,
                         m=self.mn[0], n=self.mn[1])
            Ht4 = self.restrict_coeff(Ht3, engine, vec=True, idx=mm._sel_index)
            intg = mfem.VectorFEBoundaryTangentLFIntegrator(Ht4)
            lf1i.AddBoundaryIntegrator(intg)
            lf1i.Assemble()
            v1 = LF2PyVec(lf1, lf1i)
            vecs.append(v1)

        if ext_drive is not None:
            for i in range(len(ext_drive)):
                vecs.append(vecs[-1]*0)

        t1 = HStackPyVec(vecs)
        t3 = IdentityPyMat(l, diag=-1)

        t4 = np.zeros(l)
        if Sright.shape[1] > 0:
            t4 = Sright.dot(np.atleast_2d(ext_drive).transpose())

        t4 = Array2PyVec(t4)
        # return (v1, v2, t3, t4, True)
        return (t1, None, t3, t4, True)

    def has_extra_coupling(self):
        '''
        True if it define coupling between Lagrange multipliers
        '''
        name1 = self.extra_DoF_name()

        ports = self. _get_coupled_ports()
        names2 = [x[2] for x in ports]

        if len(names2) == 0:
            return False
        return True

    def extra_coupling_names(self):
        '''
        return its own extra_name, and paired (coupled) extra_names
        '''
        name1 = self.extra_DoF_name()

        ports = self. _get_coupled_ports()
        names2 = [x[2] for x in ports]

        return name1, names2

    def get_extra_coupling(self, target_name):
        '''
        [    t2][paired extra]
        [t1    ][extra]
        t1 = (size of extra, size of targert_extra, )
        t2 = (size of target_extra, size of extra, )        
        '''
        Smat, port_idx, ext_drive = self.vt.make_value_or_expression(self)
        ports = self. _get_coupled_ports()
        names2 = [x[2] for x in ports]

        l = self.get_extra_NDoF()
        if len(np.atleast_1d(Smat)) == 1:
            Smat = np.eye(l)*Smat

        Sleft = Smat[:, :len(names2)]
        idx = names2.index(target_name)

        c1 = Sleft[:, [idx]]
        c2 = np.zeros((1, l))
        c2[0, idx] = -1

        from petram.helper.densemat2pymat import Densemat2PyMat
        ret1 = Densemat2PyMat(c1)
        ret2 = Densemat2PyMat(c2)

        return ret1, ret2



#
#  common routine for EM Physics Modules
#
from petram.phys.phys_model import PhysModule
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EMPhysModule')


class EMPhysModule(PhysModule):
    def __init__(self, **kwargs):
        super(EMPhysModule, self).__init__(**kwargs)
        self.freq_txt = "1.0e9"

    def attribute_set(self, v):
        v = super(EMPhysModule, self).attribute_set(v)
        v["freq_txt"] = "1.0e9"
        v["coeffnorm_txt"] = "1.0"
        return v

    def is_complex(self):
        return True

    def get_default_ns(self):
        from petram.phys.phys_const import mu0, epsilon0, q0, c, massu, chargez
        ns = {'mu0': mu0,
              'e0': epsilon0,
              'q0': q0,
              'clight': c,
              'massu': massu,
              'chargez': chargez}
        return ns

    def get_freq_omega(self):
        freq, _void = self.eval_param_expr(self.freq_txt, "freq")
        try:
            _void = float(freq)
        except:
            freq = 1e6
            dprint1("Error, frequency must be a scalr real value")
        return freq, 2.*np.pi*freq

    def panel1_param(self):
        panels = super(EMPhysModule, self).panel1_param()
        panels.extend([self.make_param_panel('freq',  self.freq_txt),
                       self.make_param_panel('row scaling',  self.coeffnorm_txt), ])

        return panels

    def get_panel1_value(self):
        val = super(EMPhysModule, self).get_panel1_value()
        val.extend([self.freq_txt, self.coeffnorm_txt])

        return val

    def import_panel1_value(self, v):
        v = super(EMPhysModule, self).import_panel1_value(v)
        self.freq_txt = str(v[0])
        self.coeffnorm_txt = str(v[1])
        return v[2:]

    def get_coeff_norm(self):
        g = self._global_ns.copy()
        '''
        fname = 'freq' + self.dep_vars_suffix
        f, w = self.get_freq_omega()
        if fname not in g:
            g[fname] = f
        '''
        try:
            norm = float(eval(self.coeffnorm_txt, g))
        except:
            norm = 1.0
            dprint1("Error, Coefficeint normalization be a scalr real value")
        return norm

    def add_variables(self, v, name, solr, soli=None):
        from petram.helper.variables import add_constant

        if name == self.dep_vars[0]:
            suffix = self.dep_vars_suffix

            freq, omega = self.get_freq_omega()
            add_constant(v, 'freq', suffix, freq)
            add_constant(v, 'omega', suffix, np.float64(omega))

            norm = self.get_coeff_norm()
            add_constant(v, 'rowscale', suffix, norm)

        return v



'''

   3D port boundary condition


    2016 5/20  first version only TE modes
'''
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
from petram.phys.phys_const import epsilon0, mu0
from mfem.common.mpi_debug import nicePrint
import sys
import numpy as np

from petram.model import Bdry
from petram.phys.phys_model import Phys
from petram.phys.em2da.em2da_base import EM2Da_Bdry, EM2Da_Domain

from petram.helper.geom import connect_pairs2
from petram.helper.geom import find_circle_center_radius

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM3D_Port')


if use_parallel:
    import mfem.par as mfem
    '''
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   '''
else:
    import mfem.ser as mfem

'''
  TE Mode (Ephi only)
    Expression are based on Microwave Engineering p122 - p123.
    Note that it consists from two terms
       1)  \int dS W \dot n \times iwH (VectorFETangentIntegrator does this)
       2)  an weighting to evaulate mode amplutude from the E field
           on a boundary


'''

data = (('inc_amp', VtableElement('inc_amp', type='complex',
                                  guilabel='incoming amp',
                                  default=1.0,
                                  tip="amplitude of incoming wave")),
        ('inc_phase', VtableElement('inc_phase', type='float',
                                    guilabel='incoming phase (deg)',
                                    default=0.0,
                                    tip="phase of incoming wave")),
        ('epsilonr', VtableElement('epsilonr', type='complex',
                                   guilabel='epsilonr',
                                   default=1.0,
                                   tip="relative permittivity")),
        ('mur', VtableElement('mur', type='complex',
                              guilabel='mur',
                              default=1.0,
                              tip="relative permeability")),)

'''
   E_(name)_rz  :
   E_(name)_phi :
   H_(name)_rz  : Vector (Hphi, Hphi)
   H_(name)_phi : Scalar (nr Hz - nz Hr)

   n x H  = (-nz * Hphi, nr*Hz - nz*Hr, nr * Hphi)
'''
'''
   rectangular wg TE
'''


class E_TE_phi(mfem.PyCoefficient):
    def __init__(self, bdry, real=True, amp=1.0, eps=1.0, mur=1.0):

        self.real = real
        freq, omega = bdry.get_root_phys().get_freq_omega()

        self.a, self.c = bdry.a, bdry.c
        self.a_vec = bdry.a_vec
        self.m, self.n = bdry.mn[0], bdry.mn[1]
        mfem.PyCoefficient.__init__(self)

    def EvalValue(self, x):
        p = np.array(x)
        # x, y is positive (first quadrant)
        xx = np.sqrt(np.sum((p - self.c)**2)) / self.a  # 0 < x < 1
        Ephi = self.m / self.a * \
            np.sin(self.m * np.pi * xx) * (x[0])  # Ephi = -rho Ephi
        if self.real:
            return -Ephi.real
        else:
            return -Ephi.imag


class H_TE_rz(mfem.VectorPyCoefficient):
    def __init__(
            self,
            sdim,
            phase,
            bdry,
            real=True,
            amp=1.0,
            eps=1.0,
            mur=1.0):
        mfem.VectorPyCoefficient.__init__(self, sdim)

        self.real = real
        self.phase = phase  # phase !=0 for incoming wave

        freq, omega = bdry.get_root_phys().get_freq_omega()
        cnorm = bdry.get_root_phys().get_coeff_norm()

        self.a, self.c = bdry.a, bdry.c
        self.a_vec = bdry.a_vec
        self.m, self.n = bdry.mn[0], bdry.mn[1]

        k = omega * np.sqrt(eps * epsilon0 * mur * mu0)
        kc = np.abs(bdry.mn[0] * np.pi / bdry.a)
        if kc > k:
            raise ValueError('Mode does not propagate')
        beta = np.sqrt(k**2 - kc**2)
        dprint1("propagation constant:" + str(beta))

        AA = omega * mur * mu0 * np.pi / kc / kc
        self.AA = omega * beta * np.pi / kc / kc / AA * amp * cnorm

    def EvalValue(self, x):
        p = np.array(x)
        # x, y is positive (first quadrant)
        xx = np.sqrt(np.sum((p - self.c)**2)) / self.a  # 0 < x < 1
        H = 1j * self.AA * self.m / self.a * np.sin(self.m * np.pi * xx)
        H = -H * np.exp(1j * self.phase / 180. * np.pi) * self.a_vec * x[0]
        H = np.array([H[0], 0])
        # return np.array([1,1.])
        if self.real:
            return H.real
        else:
            return H.imag


class H_TE_phi(mfem.PyCoefficient):
    def __init__(self, phase, bdry, real=True, amp=1.0, eps=1.0, mur=1.0):
        mfem.PyCoefficient.__init__(self)

        self.real = real
        self.phase = phase  # phase !=0 for incoming wave

        freq, omega = bdry.get_root_phys().get_freq_omega()
        cnorm = bdry.get_root_phys().get_coeff_norm()

        self.a, self.c = bdry.a, bdry.c
        self.a_vec = bdry.a_vec
        self.m, self.n = bdry.mn[0], bdry.mn[1]

        k = omega * np.sqrt(eps * epsilon0 * mur * mu0)
        kc = np.abs(bdry.mn[0] * np.pi / bdry.a)
        if kc > k:
            raise ValueError('Mode does not propagate')
        beta = np.sqrt(k**2 - kc**2)
        dprint1("propagation constant:" + str(beta))

        AA = omega * mur * mu0 * np.pi / kc / kc * amp
        self.AA = omega * beta * np.pi / kc / kc / AA * cnorm
        self.beta = beta
        self.fac = 1 / mu0 / mur * amp
        '''
        AA = beta/amp/mu
       '''

    def EvalValue(self, x):
        p = np.array(x)
        # x, y is positive (first quadrant)
        xx = np.sqrt(np.sum((p - self.c)**2)) / self.a  # 0 < x < 1
        #H = 1j*self.AA* self.m/self.a*np.sin(self.m*np.pi*xx)
        H = (-1j * self.beta - 1 / x[0]) * self.fac * \
            self.m / self.a * np.sin(self.m * np.pi * xx)
        H = H * np.exp(1j * self.phase / 180. * np.pi)
        if self.real:
            return H.real
        else:
            return H.imag


H_TE_rz = None
E_TE_rz = None


# Ephi mode
class E_Ephi_rz(mfem.VectorPyCoefficient):
    pass


class E_Ephi_phi(mfem.PyCoefficient):
    pass


class H_Ephi_rz(mfem.VectorPyCoefficient):
    pass


class H_Ephi_phi(mfem.PyCoefficient):
    pass


def bdry_constraints():
    return [EM2Da_Port]


class EM2Da_Port(EM2Da_Bdry):
    extra_diagnostic_print = True
    vt = Vtable(data)

    def __init__(self, mode='TE', mn='0,1', inc_amp='1',
                 inc_phase='0', port_idx=1):
        super(EM2Da_Port, self).__init__(mode=mode,
                                         mn=mn,
                                         inc_amp=inc_amp,
                                         inc_phase=inc_phase,
                                         port_idx=port_idx)
        Phys.__init__(self)

    def attribute_set(self, v):
        super(EM2Da_Port, self).attribute_set(v)
        v['port_idx'] = 1
        v['mode'] = 'TE'
        v['mn'] = [1, 0]
        v['epsilonr'] = 1.0
        v['mur'] = 1.0
        v['sel_readonly'] = False
        v['sel_index'] = []
        v['isTimeDependent_RHS'] = True
        self.vt.attribute_set(v)
        return v

    def panel1_param(self):
        return ([["port id", str(self.port_idx), 0, {}],
                 ["mode (n=0 only)", self.mode, 4, {"readonly": True,
                                                    "choices": ["TE", ]}],
                 ["m/n", ','.join(str(x) for x in self.mn), 0, {}], ] +
                self.vt.panel_param(self))

    def get_panel1_value(self):
        return ([str(self.port_idx),
                 self.mode, ','.join(str(x) for x in self.mn)] +
                self.vt.get_panel_value(self))

    def import_panel1_value(self, v):
        self.port_idx = v[0]
        self.mode = v[1]
        self.mn = [int(x) for x in v[2].split(',')]
        self.vt.import_panel_value(self, v[3:])

    def panel4_param(self):
        ll = super(EM2Da_Port, self).panel4_param()
        ll.append(['Varying (in time/for loop) RHS', False, 3, {"text": ""}])
        return ll

    def panel4_tip(self):
        return None

    def import_panel4_value(self, value):
        super(EM2Da_Port, self).import_panel4_value(value[:-1])
        self.isTimeDependent_RHS = value[-1]

    def get_panel4_value(self):
        value = super(EM2Da_Port, self).get_panel4_value()
        value.append(self.isTimeDependent_RHS)
        return value

    def verify_setting(self):
        if self.isTimeDependent_RHS:
            flag = True
        else:
            flag = False
        return flag, 'Varying RHS is not set', 'This potntially causes an error with PortScan. Set it Time/NL Dep. panel '

    def update_param(self):
        self.update_inc_amp_phase()

    def update_inc_amp_phase(self):
        '''
        set parameter
        '''
        try:
            self.inc_amp = self.eval_phys_expr(
                str(self.inc_amp_txt), 'inc_amp')[0]
            self.inc_phase = self.eval_phys_expr(
                str(self.inc_phase_txt), 'inc_phase', chk_float=True)[0]
        except BaseException:
            raise ValueError("Cannot evaluate amplitude/phase to float number")

    def preprocess_params(self, engine):
        # find normal (outward) vector...
        mesh = engine.get_emesh(mm=self)
        fespace = engine.fespaces[self.get_root_phys().dep_vars[0]]
        nbe = mesh.GetNBE()
        ibe = np.array([i for i in range(nbe)
                        if mesh.GetBdrElement(i).GetAttribute() ==
                        self._sel_index[0]])
        dprint1("idb", ibe)
        el = mesh.GetBdrElement(ibe[0])
        Tr = fespace.GetBdrElementTransformation(ibe[0])
        rules = mfem.IntegrationRules()
        ir = rules.Get(el.GetGeometryType(), 1)
        Tr.SetIntPoint(ir.IntPoint(0))
        nor = mfem.Vector(2)
        mfem.CalcOrtho(Tr.Jacobian(), nor)

        self.norm = nor.GetDataArray().copy()
        self.norm = self.norm / np.sqrt(np.sum(self.norm**2))

        dprint1("Normal Vector " + list(self.norm).__repr__())

        # find rectangular shape

        edges = np.array([mesh.GetBdrElementEdges(i)[0]
                          for i in ibe]).flatten()
        d = {}
        for x in edges:
            d[x] = x in d
        edges = [x for x in d.keys() if not d[x]]
        ivert = [mesh.GetEdgeVertices(x) for x in edges]
        ivert = connect_pairs2(ivert)[0]

        vv = np.vstack([mesh.GetVertexArray(i) for i in ivert])
        self.ctr = (vv[0] + vv[-1]) / 2.0
        dprint1("Center " + list(self.ctr).__repr__())

    # rectangular port
        self.a_vec = (vv[0] - vv[-1])
        self.a = np.sqrt(np.sum(self.a_vec**2))
        self.a_vec /= self.a
        self.c = vv[-1]
        dprint1("Cornor " + self.c.__repr__())
        dprint1("Edge  " + self.a.__repr__())
        dprint1("Edge Vec." + list(self.a_vec).__repr__())

        Erz, Ephi = self.get_e_coeff_cls()
        Hrz, Hphi = self.get_h_coeff_cls()

        if self.mode == 'TE':
            dprint1("E field pattern")
            c1 = Ephi(self, real=True)
            c2 = Hphi(0.0, self, real=False)
            for p in vv:
                dprint1(p.__repr__() + ' : ' + c1.EvalValue(p).__repr__())
            dprint1("H field pattern")
            for p in vv:
                dprint1(p.__repr__() + ' : ' + c2.EvalValue(p).__repr__())

    def get_h_coeff_cls(self):
        m = sys.modules[__name__]
        c1 = getattr(m, 'H_' + self.mode + '_rz')
        c2 = getattr(m, 'H_' + self.mode + '_phi')
        return c1, c2

    def get_e_coeff_cls(self):
        m = sys.modules[__name__]
        c1 = getattr(m, 'E_' + self.mode + '_rz')
        c2 = getattr(m, 'E_' + self.mode + '_phi')
        return c1, c2

    def has_lf_contribution(self, kfes):
        self.vt.preprocess_params(self)
        inc_amp, inc_phase, eps, mur = self.vt.make_value_or_expression(self)
        if kfes == 1:
            if self.mode == 'TE':
                return inc_amp != 0
            elif self.mode == 'Ephi':
                return inc_amp != 0
        elif kfes == 0:
            if self.mode == 'TE':
                return inc_amp != 0
            elif self.mode == 'Ephi':
                return False
        else:
            return False

    def add_lf_contribution(self, engine, b, real=True, kfes=0):
        if real:
            dprint1("Add LF contribution(real)" + str(self._sel_index))
        else:
            dprint1("Add LF contribution(imag)" + str(self._sel_index))

        self.vt.preprocess_params(self)
        inc_amp, inc_phase, eps, mur = self.vt.make_value_or_expression(self)

        Hrz, Hphi = self.get_h_coeff_cls()
        inc_wave = inc_amp * np.exp(1j * inc_phase / 180. * np.pi)
        # inc_wave = inc_amp * np.exp(-1j*inc_phase/180.*np.pi)  # original...
        #assert False, "when you use for the first time, the sign must be checked...)"
        phase = np.angle(inc_wave) * 180 / np.pi
        amp = np.abs(inc_wave)

#        if (kfes == 0 and self.mode == 'TE'):  # E~phi B ~rz
#            coeff = Hrz(2, phase, self, real = real, amp = amp)
#            self.add_integrator(engine, 'inc_amp', coeff,
#                                b.AddBoundaryIntegrator,
#                                mfem.VectorFEDomainLFIntegrator)
#                                mfem.VectorFEBoundaryTangentLFIntegrator)
        if (kfes == 1 and self.mode == 'TE'):  # E~phi B ~rz
            coeff = Hphi(phase, self, real=real, amp=amp, eps=eps, mur=mur)
            self.add_integrator(engine, 'inc_amp', coeff,
                                b.AddBoundaryIntegrator,
                                mfem.BoundaryLFIntegrator)

        else:
            pass

    def has_extra_DoF(self, kfes):
        if self.mode == 'TE' and kfes == 1:
            return True
        return False

    def get_extra_NDoF(self):
        return 1

    def extra_DoF_name2(self, kfes=0):
        '''
        default DoF name
        '''
        if self.mode == 'TE' and kfes == 1:
            return self.get_root_phys().dep_vars[1]+"_port"+str(self.port_idx)

        assert False, "should not come here"

    def get_probes(self):
        if self.mode == 'TE':
            return [self.get_root_phys().dep_vars[1]+"_port"+str(self.port_idx), ]
        else:
            return []

    def postprocess_extra(self, sol, flag, sol_extra):
        name = self.name() + '_' + str(self.port_idx)
        sol_extra[name] = sol.toarray()

    def add_extra_contribution(self, engine, **kwargs):
        from mfem.common.chypre import LF2PyVec

        kfes = kwargs.pop('kfes', 0)
        dprint1("Add Extra contribution" + str(self._sel_index))

        self.vt.preprocess_params(self)
        inc_amp, inc_phase, eps, mur = self.vt.make_value_or_expression(self)

        Erz, Ephi = self.get_e_coeff_cls()
        Hrz, Hphi = self.get_h_coeff_cls()
        fes = engine.get_fes(self.get_root_phys(), kfes)

        '''
        if (kfes == 0 and self.mode == 'TE'):
           lf1 = engine.new_lf(fes)
           Ht = Hrz(2, 0.0, self, real = True)
           Ht = self.restrict_coeff(Ht, engine, vec = True)
           intg =  mfem.VectorFEDomainLFIntegrator(Ht)
           #intg = mfem.VectorFEBoundaryTangentLFIntegrator(Ht)
           lf1.AddBoundaryIntegrator(intg)
           lf1.Assemble()
           lf1i = engine.new_lf(fes)
           Ht = Hrz(2, 0.0, self, real = False)
           Ht = self.restrict_coeff(Ht, engine, vec = True)
           intg =  mfem.VectorFEDomainLFIntegrator(Ht)
           #intg = mfem.VectorFEBoundaryTangentLFIntegrator(Ht)
           lf1i.AddBoundaryIntegrator(intg)
           lf1i.Assemble()

           from mfem.common.chypre import LF2PyVec
           v1 = LF2PyVec(lf1, lf1i)
           v1 *= -1
           # output formats of InnerProduct
           # are slightly different in parallel and serial
           # in serial numpy returns (1,1) array, while in parallel
           # MFEM returns a number. np.sum takes care of this.
           return (v1, None, None, None, False)
        '''
        if (kfes == 1 and self.mode == 'TE'):

            lf1 = engine.new_lf(fes)
            Ht = Hphi(0.0, self, real=True, eps=eps, mur=mur)
            Ht = self.restrict_coeff(Ht, engine)
            intg = mfem.BoundaryLFIntegrator(Ht)
            lf1.AddBoundaryIntegrator(intg)
            lf1.Assemble()

            lf1i = engine.new_lf(fes)
            Ht = Hphi(0.0, self, real=False, eps=eps, mur=mur)
            Ht = self.restrict_coeff(Ht, engine)
            intg = mfem.BoundaryLFIntegrator(Ht)
            lf1i.AddBoundaryIntegrator(intg)
            lf1i.Assemble()

            from mfem.common.chypre import LF2PyVec, PyVec2PyMat, Array2PyVec, IdentityPyMat

            v1 = LF2PyVec(lf1, lf1i)
            #v1 *= -1

            lf2 = engine.new_lf(fes)
            Et = Ephi(self, real=True, eps=eps, mur=mur)
            Et = self.restrict_coeff(Et, engine)
            intg = mfem.DomainLFIntegrator(Et)
            lf2.AddBoundaryIntegrator(intg)
            lf2.Assemble()

            x = engine.new_gf(fes)
            x.Assign(0.0)
            arr = self.get_restriction_array(engine)
            x.ProjectBdrCoefficient(Et, arr)

            weight = mfem.InnerProduct(engine.x2X(x), engine.b2B(lf2))
            v2 = LF2PyVec(lf2, None, horizontal=True)
            v2 *= -1 / weight / 2.0

            #x  = LF2PyVec(x, None)
            # output formats of InnerProduct
            # are slightly different in parallel and serial
            # in serial numpy returns (1,1) array, while in parallel
            # MFEM returns a number. np.sum takes care of this.
            #tmp = np.sum(v2.dot(x))
            #v2 *= -1/tmp/2.

            t4 = np.array([[inc_amp * np.exp(1j * inc_phase / 180. * np.pi)]])

            # convert to a matrix form
            v1 = PyVec2PyMat(v1)
            v2 = PyVec2PyMat(v2.transpose())
            t4 = Array2PyVec(t4)
            t3 = IdentityPyMat(1, diag=-1)

            v2 = v2.transpose()

            # return (None, v2, t3, t4, True)
            return (v1, v2, t3, t4, True)



'''
  locak-K plasma
'''
from petram.phys.common.rf_dispersion_lkplasma import (vtable_data0,
                                                       default_kpe_option,
                                                       kpe_options,
                                                       term_options,
                                                       default_term_option)

from petram.phys.common.rf_dispersion_coldplasma import (col_model_options,
                                                         default_col_model,)

from petram.phys.em2d.em2d_base import EM2D_Bdry, EM2D_Domain, EM2D_Domain_helper
from petram.phys.phys_const import mu0, epsilon0
from petram.phys.vtable import VtableElement, Vtable
from petram.mfem_config import use_parallel
import numpy as np

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('EM2D_LocalK')

if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

vtable_data = vtable_data0.copy()
vtable_data.extend([('kz', VtableElement('kz', type='float',
                                         guilabel='kz',
                                         default=0.0,
                                         no_func=True,
                                         tip="out-of-plane wave number")), ])

kpe_alg_options = ["std", "em2d"]


def domain_constraints():
    return [EM2D_LocalK]


class EM2D_LocalK(EM2D_Domain, EM2D_Domain_helper):
    vt = Vtable(vtable_data)
    # nlterms = ['epsilonr']

    @classmethod
    def fancy_menu_name(cls):
        return "HotPlasma(local-k)"

    @classmethod
    def fancy_tree_name(cls):
        return "LocalKPlasma"

    def get_possible_child(self):
        from .em2d_pml import EM2D_LinearPML
        return [EM2D_LinearPML]

    def has_bf_contribution(self, kfes):
        if kfes == 0:
            return True
        elif kfes == 1:
            return True
        else:
            return False

    def has_mixed_contribution(self):
        return True

    def get_mixedbf_loc(self):
        '''
        r, c, and flag1, flag2 of MixedBilinearForm
           flag1 : take transpose
           flag2 : take conj
        '''
        return [(0, 1, 1, 1), (1, 0, 1, 1), ]  # (0, 1, -1, 1)]

    def attribute_set(self, v):
        EM2D_Domain.attribute_set(self, v)
        v["kpe_mode"] = default_kpe_option
        v["kpe_alg"] = kpe_alg_options[0]
        v["col_model"] = default_col_model
        v["lk_terms"] = None
        return v

    def config_terms(self, evt):
        from petram.phys.common.rf_lk_term_panel import ask_rf_lk_terms

        self.vt.preprocess_params(self)
        ret = self.vt.make_value_or_expression(self)
        charges = ret[7]

        num_ions = len(charges)
        win = evt.GetEventObject()
        value = ask_rf_lk_terms(win, num_ions, self.lk_terms)
        self.lk_terms = value

    def panel1_param(self):
        panels = super(EM2D_LocalK, self).panel1_param()
        panels.append(["kpe mode", None, 1, {"values": kpe_options}])
        panels.append(["kpe alg.", None, 1, {"values": kpe_alg_options}])
        panels.append(["col. model", None, 1, {"values": col_model_options}])
        panels.extend([  # ["contributions", "", 2, None],
            [None, None, 341, {"label": "Customize terms",
                               "func": "config_terms",
                               "sendevent": True,
                               "noexpand": True}], ])
        return panels

    def get_panel1_value(self):
        values = super(EM2D_LocalK, self).get_panel1_value()
        values.extend([self.kpe_mode, self.kpe_alg, self.col_model, self])
        return values

    def import_panel1_value(self, v):
        check = super(EM2D_LocalK, self).import_panel1_value(v[:-4])
        self.kpe_mode = v[-4]
        self.kpe_alg = v[-3]
        self.col_model = v[-2]
        return check

    @ property
    def jited_coeff(self):
        return self._jited_coeff

    def compile_coeffs(self):
        self._jited_coeff = self.get_coeffs()

    def get_coeffs(self):
        _freq, omega = self.get_root_phys().get_freq_omega()
        B, dens_e, t_e, dens_i, t_i, t_c, masses, charges, kpakpe, kpevec, kz = self.vt.make_value_or_expression(
            self)
        ind_vars = self.get_root_phys().ind_vars
        kpe_mode = self.kpe_mode
        kpe_alg = self.kpe_alg

        from petram.phys.common.rf_dispersion_lkplasma import value2flags

        num_ions = len(charges)
        terms = value2flags(num_ions, self.lk_terms)

        from petram.phys.common.rf_dispersion_lkplasma import build_coefficients
        coeff1, coeff2, coeff3 = build_coefficients(ind_vars, omega, B, t_c, dens_e, t_e,
                                                            dens_i, t_i, masses, charges, kpakpe, kpevec,
                                                            kpe_mode, self.col_model, cnorm,
                                                            self._global_ns, self._local_ns,
                                                            kpe_alg=kpe_alg, sdim=2, kzmode=kz, terms=terms)

        return coeff1, coeff2, coeff3, kz

    def get_coeffs_2(self):
        # e, m, s
        coeff1, coeff2, coeff3, kz = self.jited_coeff
        '''
        coeff4 = ComplexMatrixSum(
            coeff1, coeff3)      # -> coeff4 = coeff1 + coeff3
        '''
        coeff4 = coeff1 + coeff3

        if self.has_pml():
            coeff4 = self.make_PML_coeff(coeff4)
            coeff2 = self.make_PML_coeff(coeff2)

        eps11 = coeff4[[0, 1], [0, 1]]
        eps21 = coeff4[[0, 1], 2]
        eps12 = coeff4[2, [0, 1]]
        eps22 = coeff4[2, 2]

        eps = [eps11, eps12, eps21, eps22]

        tmp = coeff2.inv()

        mu11 = tmp[[0, 1], [0, 1]]
        mu11 = mu11.adj()
        mu22 = tmp[2, 2]
        k2_over_mu11 = mu11*(kz*kz)
        ik_over_mu11 = mu11*(1j*kz)

        mu = [mu11, mu22, k2_over_mu11, ik_over_mu11]

        return eps, mu, kz

    def add_bf_contribution(self, engine, a, real=True, kfes=0):
        # freq, omega = self.get_root_phys().get_freq_omega()
        eps, mu, kz = self.get_coeffs_2()

        self.set_integrator_realimag_mode(real)

        if real:
            if kfes == 0:
                dprint1("Add ND contribution(real)" + str(self._sel_index))
            elif kfes == 1:
                dprint1("Add H1 contribution(real)" + str(self._sel_index))
        else:
            if kfes == 0:
                dprint1("Add ND contribution(imag)" + str(self._sel_index))
            elif kfes == 1:
                dprint1("Add H1 contribution(imag)" + str(self._sel_index))

        self.call_bf_add_integrator(eps,  mu, kz,  engine, a, kfes)

    def add_mix_contribution(self, engine, mbf, r, c, is_trans, real=True):
        if real:
            dprint1("Add mixed contribution(real)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))
        else:
            dprint1("Add mixed contribution(imag)" + "(" + str(r) + "," + str(c) + ')'
                    + str(self._sel_index))

        # freq, omega = self.get_root_phys().get_freq_omega()
        eps, mu, kz = self.get_coeffs_2()

        self.set_integrator_realimag_mode(real)

        self.call_mix_add_integrator(eps, mu, engine, mbf, r, c, is_trans)

    def add_domain_variables(self, v, n, suffix, ind_vars):
        from petram.helper.variables import add_constant

        if len(self._sel_index) == 0:
            return

        freq, omega = self.get_root_phys().get_freq_omega()
        B, dens_e, t_e, dens_i, t_i, t_c, masses, charges, kpakpe, kpevec, kz = self.vt.make_value_or_expression(
            self)

        kpe_mode = self.kpe_mode
        kpe_alg = self.kpe_alg

        add_constant(v, 'kz', suffix, np.float64(kz),
                     domains=self._sel_index,
                     gdomain=self._global_ns)

        from petram.phys.common.rf_dispersion_lkplasma import value2flags

        num_ions = len(charges)
        terms = value2flags(num_ions, self.lk_terms)

        from petram.phys.common.rf_dispersion_lkplasma import build_variables

        ss = self.parent.parent.name()+'_'+self.name()  # phys module name + name
        ret = build_variables(v, ss, ind_vars,
                              omega, B, t_c, dens_e, t_e,
                              dens_i, t_i, masses, charges,
                              kpakpe, kpevec, kpe_mode, kpe_alg, self.col_model,
                              self._global_ns, self._local_ns,
                              sdim=2, terms=terms)

        from petram.phys.common.rf_dispersion_lkplasma import add_domain_variables_common
        add_domain_variables_common(self, ret, v, suffix, ind_vars)

        var = ['x', 'y', 'z']
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'epsilonr')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'mur')
        self.do_add_matrix_component_expr(v, suffix, ind_vars, var, 'sigma')

        return
